#include <NTL/GF2X.h>
#include <fstream>
#include <cstdio>
#include <cstdlib>

using namespace std;
#define MEXP 19937
#define N 624
#define M 397
#define MATRIX_A 0x9908b0dful   /* constant vector a */
#define MSB 0x80000000ul 
#define LSB 0x00000001ul 
#define MASK 0x80000000ul

NTL_CLIENT

struct state{
  unsigned int s[N];
  int ptr;
};

void init_state(struct state *);
unsigned int genrand_int32(struct state *);
void recover(struct state *st, unsigned int *);
int compare_state(struct state *, struct state *);
int compare_state2(struct state *s1, struct state *s2);
void init_poly(void);

GF2X pf(INIT_SIZE, MEXP);

int main(void)
{
  ifstream fin;
  int i, j;
  unsigned int x[MEXP+1], jump=1234567;
  char c;
  GF2X pg(INIT_SIZE, 2*MEXP-1), ph(INIT_SIZE, 3*MEXP-2);
  struct state mt0, mt1, mt;

  fin.open("clist_mt19937.txt");
  if (!fin){
    cout << "File read error.\n";
    return 1;
  }

  for (i=MEXP-1; i>-1; i--){
    fin >> c;
    SetCoeff(pf, i, c);
  }
    
  genrand_int32(&mt);
  
  for (i=2*MEXP-2; i>-1; i--)
    SetCoeff(pg, i, genrand_int32(&mt));
  
  ph = pf * pg;
  
  for (i=0; i<MEXP; i++)
    x[i+1] = IsOne(coeff(ph,2*MEXP-2-i)); 
  
  recover (&mt1, x);
  
  return 0;
}

void init_poly(void)
{
  int i;

  for (i=0; i<MEXP; i++)
    SetCoeff(pf, i, rand() & LSB);
}

void init_state(struct state *st)
{
  int i;

  st->ptr = 0;

  for (i=0; i<N; i++)
    st->s[i] = (rand() << 1) | rand();
}

unsigned int genrand_int32 (struct state *st)
{

  unsigned int y;
  static unsigned long mag01[2]={0x0UL, MATRIX_A};

  if (st->ptr >= N){
    st->ptr = 0;
    int kk;
    for (kk=0;kk<N-M;kk++) {
      y = (st->s[kk]&MASK)|(st->s[kk+1]&~MASK);
      st->s[kk] = st->s[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
    }
    for (;kk<N-1;kk++) {
      y = (st->s[kk]&MASK)|(st->s[kk+1]&~MASK);
      st->s[kk] = st->s[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
    }
    y = (st->s[N-1]&MASK)|(st->s[0]&~MASK);
    st->s[N-1] = st->s[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];
  }

  y = st->s[st->ptr];
  st->ptr++;

  return y;
}

int compare_state(struct state *s1, struct state *s2)
{
  int i;

  if ((s1->s[s1->ptr] & MASK) != (s2->s[s2->ptr] & MASK))
    return 0;

  for (i=1; i<N; i++){
    if ((s1->s[(s1->ptr+i)%N]) != (s2->s[(s2->ptr+i)%N]))
      return 0;
    else 
      ;
  }

  return 1;
}

int compare_state2(struct state *s1, struct state *s2)
{
  int i;
  
  genrand_int32(s1);
  genrand_int32(s2);

  for (i=0; i<MEXP; i++){
    if ((genrand_int32(s1) & LSB) != (genrand_int32(s2) & LSB))
      return 0;
  }

  return 1;
}


void recover(struct state *st, unsigned int x[MEXP+1])
{
  int i;
  unsigned int y;

  for (i=MEXP; i>N-1; i--){
    y = x[i] ^ x[i-N+M];
    if ((x[i-N+1] & LSB) != 0)
      y ^= MATRIX_A;
    else
      ;
    y = (y << 1);
    y |= (LSB & x[i-N+1]);
    x[i-N+1] = (MASK & x[i-N+1]) ^ (~MASK & y);
    x[i-N] = (MASK & y) ^ (~MASK & x[i-N]);
  }

  for (i=0; i<N; i++)
    st->s[i] = x[i];
  st->ptr = 0;

}
